# coding: utf-8
"""api for working with tasks"""

# docs
from typing import List, NamedTuple, Dict, Optional, Callable
import os
import time
from collections import defaultdict, OrderedDict
import json

from supervisely.api.module_api import (
    ApiField,
    ModuleApiBase,
    ModuleWithStatus,
    WaitingTimeExceeded,
)
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from supervisely.io.fs import get_file_name, ensure_base_path, get_file_hash
from supervisely.collection.str_enum import StrEnum
from supervisely._utils import batched, take_with_default


class TaskFinishedWithError(Exception):
    """TaskFinishedWithError"""

    pass


class TaskApi(ModuleApiBase, ModuleWithStatus):
    """
    API for working with Tasks. :class:`TaskApi<TaskApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        task_id = 121230
        task_info = api.task.get_info_by_id(task_id)
    """

    class RestartPolicy(StrEnum):
        """RestartPolicy"""

        NEVER = "never"
        """"""
        ON_ERROR = "on_error"
        """"""

    class PluginTaskType(StrEnum):
        """PluginTaskType"""

        TRAIN = "train"
        """"""
        INFERENCE = "inference"
        """"""
        INFERENCE_RPC = "inference_rpc"
        """"""
        SMART_TOOL = "smarttool"
        """"""
        CUSTOM = "custom"
        """"""

    class Status(StrEnum):
        """Status"""

        QUEUED = "queued"
        """"""
        CONSUMED = "consumed"
        """"""
        STARTED = "started"
        """"""
        DEPLOYED = "deployed"
        """"""
        ERROR = "error"
        """"""
        FINISHED = "finished"
        """"""
        TERMINATING = "terminating"
        """"""
        STOPPED = "stopped"
        """"""

    def __init__(self, api):
        ModuleApiBase.__init__(self, api)
        ModuleWithStatus.__init__(self)

    def get_list(
        self, workspace_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        List of Tasks in the given Workspace.

        :param workspace_id: Workspace ID.
        :type workspace_id: int
        :param filters: List of params to sort output Projects.
        :type filters: List[dict], optional
        :return: List of Tasks with information for the given Workspace.
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            workspace_id = 23821

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            task_infos = api.task.get_list(workspace_id)

            task_infos_filter = api.task.get_list(23821, filters=[{'field': 'id', 'operator': '=', 'value': 121230}])
            print(task_infos_filter)
            # Output: [
            #     {
            #         "id": 121230,
            #         "type": "clone",
            #         "status": "finished",
            #         "startedAt": "2019-12-19T12:13:09.702Z",
            #         "finishedAt": "2019-12-19T12:13:09.701Z",
            #         "meta": {
            #             "input": {
            #                 "model": {
            #                     "id": 1849
            #                 },
            #                 "isExternal": true,
            #                 "pluginVersionId": 84479
            #             },
            #             "output": {
            #                 "model": {
            #                     "id": 12380
            #                 },
            #                 "pluginVersionId": 84479
            #             }
            #         },
            #         "description": ""
            #     }
            # ]
        """
        return self.get_list_all_pages(
            "tasks.list",
            {ApiField.WORKSPACE_ID: workspace_id, ApiField.FILTER: filters or []},
        )

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Task information by ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :return: Information about Task.
        :rtype: :class:`NamedTuple`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            task_id = 121230

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            task_info = api.task.get_info_by_id(task_id)
            print(task_info)
            # Output: {
            #     "id": 121230,
            #     "workspaceId": 23821,
            #     "description": "",
            #     "type": "clone",
            #     "status": "finished",
            #     "startedAt": "2019-12-19T12:13:09.702Z",
            #     "finishedAt": "2019-12-19T12:13:09.701Z",
            #     "userId": 16154,
            #     "meta": {
            #         "input": {
            #             "model": {
            #                 "id": 1849
            #             },
            #             "isExternal": true,
            #             "pluginVersionId": 84479
            #         },
            #         "output": {
            #             "model": {
            #                 "id": 12380
            #             },
            #             "pluginVersionId": 84479
            #         }
            #     },
            #     "settings": {},
            #     "agentName": null,
            #     "userLogin": "alexxx",
            #     "teamId": 16087,
            #     "agentId": null
            # }
        """
        return self._get_info_by_id(id, "tasks.info")

    def get_status(self, task_id: int) -> Status:
        """
        Check status of Task by ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :return: Status object
        :rtype: :class:`Status`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            task_id = 121230

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            task_status = api.task.get_status(task_id)
            print(task_status)
            # Output: finished
        """
        status_str = self.get_info_by_id(task_id)[ApiField.STATUS]  # @TODO: convert json to tuple
        return self.Status(status_str)

    def raise_for_status(self, status: Status) -> None:
        """
        Raise error if Task status is ERROR.

        :param status: Status object.
        :type status: Status
        :return: None
        :rtype: :class:`NoneType`
        """
        if status is self.Status.ERROR:
            raise TaskFinishedWithError(f"Task finished with status {str(self.Status.ERROR)}")

    def wait(
        self,
        id: int,
        target_status: Status,
        wait_attempts: Optional[int] = None,
        wait_attempt_timeout_sec: Optional[int] = None,
    ):
        """
        Awaiting achievement by given Task of a given status.

        :param id: Task ID in Supervisely.
        :type id: int
        :param target_status: Status object(status of task we expect to destinate).
        :type target_status: Status
        :param wait_attempts: The number of attempts to determine the status of the task that we are waiting for.
        :type wait_attempts: int, optional
        :param wait_attempt_timeout_sec: Number of seconds for intervals between attempts(raise error if waiting time exceeded).
        :type wait_attempt_timeout_sec: int, optional
        :return: True if the desired status is reached, False otherwise
        :rtype: :class:`bool`
        """
        wait_attempts = wait_attempts or self.MAX_WAIT_ATTEMPTS
        effective_wait_timeout = wait_attempt_timeout_sec or self.WAIT_ATTEMPT_TIMEOUT_SEC
        for attempt in range(wait_attempts):
            status = self.get_status(id)
            self.raise_for_status(status)
            if status in [
                target_status,
                self.Status.FINISHED,
                self.Status.DEPLOYED,
                self.Status.STOPPED,
            ]:
                return
            time.sleep(effective_wait_timeout)
        raise WaitingTimeExceeded(
            f"Waiting time exceeded: total waiting time {wait_attempts * effective_wait_timeout} seconds, i.e. {wait_attempts} attempts for {effective_wait_timeout} seconds each"
        )

    def upload_dtl_archive(
        self, task_id: int, archive_path: str, progress_cb: Optional[Callable] = None
    ):
        """upload_dtl_archive"""
        encoder = MultipartEncoder(
            {
                "id": str(task_id).encode("utf-8"),
                "name": get_file_name(archive_path),
                "archive": (
                    os.path.basename(archive_path),
                    open(archive_path, "rb"),
                    "application/x-tar",
                ),
            }
        )

        def callback(monitor_instance):
            read_mb = monitor_instance.bytes_read / 1024.0 / 1024.0
            if progress_cb is not None:
                progress_cb(read_mb)

        monitor = MultipartEncoderMonitor(encoder, callback)
        self._api.post("tasks.upload.dtl_archive", monitor)

    def _deploy_model(
        self,
        agent_id,
        model_id,
        plugin_id=None,
        version=None,
        restart_policy=RestartPolicy.NEVER,
        settings=None,
    ):
        """_deploy_model"""
        response = self._api.post(
            "tasks.run.deploy",
            {
                ApiField.AGENT_ID: agent_id,
                ApiField.MODEL_ID: model_id,
                ApiField.RESTART_POLICY: restart_policy.value,
                ApiField.SETTINGS: settings or {"gpu_device": 0},
                ApiField.PLUGIN_ID: plugin_id,
                ApiField.VERSION: version,
            },
        )
        return response.json()[ApiField.TASK_ID]

    def get_context(self, id: int) -> Dict:
        """
        Get context information by task ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :return: Context information in dict format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            task_id = 121230

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            context = api.task.get_context(task_id)
            print(context)
            # Output: {
            #     "team": {
            #         "id": 16087,
            #         "name": "alexxx"
            #     },
            #     "workspace": {
            #         "id": 23821,
            #         "name": "my_super_workspace"
            #     }
            # }
        """
        response = self._api.post("GetTaskContext", {ApiField.ID: id})
        return response.json()

    def _convert_json_info(self, info: dict):
        """_convert_json_info"""
        return info

    def run_dtl(self, workspace_id: int, dtl_graph: Dict, agent_id: Optional[int] = None):
        """run_dtl"""
        response = self._api.post(
            "tasks.run.dtl",
            {
                ApiField.WORKSPACE_ID: workspace_id,
                ApiField.CONFIG: dtl_graph,
                "advanced": {ApiField.AGENT_ID: agent_id},
            },
        )
        return response.json()[ApiField.TASK_ID]

    def _run_plugin_task(
        self,
        task_type,
        agent_id,
        plugin_id,
        version,
        config,
        input_projects,
        input_models,
        result_name,
    ):
        """_run_plugin_task"""
        response = self._api.post(
            "tasks.run.plugin",
            {
                "taskType": task_type,
                ApiField.AGENT_ID: agent_id,
                ApiField.PLUGIN_ID: plugin_id,
                ApiField.VERSION: version,
                ApiField.CONFIG: config,
                "projects": input_projects,
                "models": input_models,
                ApiField.NAME: result_name,
            },
        )
        return response.json()[ApiField.TASK_ID]

    def run_train(
        self,
        agent_id: int,
        input_project_id: int,
        input_model_id: int,
        result_nn_name: str,
        train_config: Optional[Dict] = None,
    ):
        """run_train"""
        model_info = self._api.model.get_info_by_id(input_model_id)
        return self._run_plugin_task(
            task_type=TaskApi.PluginTaskType.TRAIN.value,
            agent_id=agent_id,
            plugin_id=model_info.plugin_id,
            version=None,
            input_projects=[input_project_id],
            input_models=[input_model_id],
            result_name=result_nn_name,
            config={} if train_config is None else train_config,
        )

    def run_inference(
        self,
        agent_id: int,
        input_project_id: int,
        input_model_id: int,
        result_project_name: str,
        inference_config: Optional[Dict] = None,
    ):
        """run_inference"""
        model_info = self._api.model.get_info_by_id(input_model_id)
        return self._run_plugin_task(
            task_type=TaskApi.PluginTaskType.INFERENCE.value,
            agent_id=agent_id,
            plugin_id=model_info.plugin_id,
            version=None,
            input_projects=[input_project_id],
            input_models=[input_model_id],
            result_name=result_project_name,
            config={} if inference_config is None else inference_config,
        )

    def get_training_metrics(self, task_id: int):
        """get_training_metrics"""
        response = self._get_response_by_id(
            id=task_id, method="tasks.train-metrics", id_field=ApiField.TASK_ID
        )
        return response.json() if (response is not None) else None

    def deploy_model(self, agent_id: int, model_id: int) -> int:
        """deploy_model"""
        task_ids = self._api.model.get_deploy_tasks(model_id)
        if len(task_ids) == 0:
            task_id = self._deploy_model(agent_id, model_id)
        else:
            task_id = task_ids[0]
        self.wait(task_id, self.Status.DEPLOYED)
        return task_id

    def deploy_model_async(self, agent_id: int, model_id: int) -> int:
        """deploy_model_async"""
        task_ids = self._api.model.get_deploy_tasks(model_id)
        if len(task_ids) == 0:
            task_id = self._deploy_model(agent_id, model_id)
        else:
            task_id = task_ids[0]
        return task_id

    def start(
        self,
        agent_id,
        app_id=None,
        workspace_id=None,
        description="application description",
        params=None,
        log_level="info",
        users_ids=None,
        app_version="",
        is_branch=False,
        task_name="pythonSpawned",
        restart_policy="never",
        proxy_keep_url=False,
        module_id=None,
        redirect_requests={},
    ):
        """start"""
        if app_id is not None and module_id is not None:
            raise ValueError("Only one of the arguments (app_id or module_id) have to be defined")
        if app_id is None and module_id is None:
            raise ValueError("One of the arguments (app_id or module_id) have to be defined")

        data = {
            ApiField.AGENT_ID: agent_id,
            # "nodeId": agent_id,
            ApiField.WORKSPACE_ID: workspace_id,
            ApiField.DESCRIPTION: description,
            ApiField.PARAMS: take_with_default(params, {"state": {}}),
            ApiField.LOG_LEVEL: log_level,
            ApiField.USERS_IDS: take_with_default(users_ids, []),
            ApiField.APP_VERSION: app_version,
            ApiField.IS_BRANCH: is_branch,
            ApiField.TASK_NAME: task_name,
            ApiField.RESTART_POLICY: restart_policy,
            ApiField.PROXY_KEEP_URL: proxy_keep_url,
        }
        if len(redirect_requests) > 0:
            data[ApiField.REDIRECT_REQUESTS] = redirect_requests

        if app_id is not None:
            data[ApiField.APP_ID] = app_id
        if module_id is not None:
            data[ApiField.MODULE_ID] = module_id
        resp = self._api.post(method="tasks.run.app", data=data)
        task = resp.json()[0]
        if "id" not in task:
            task["id"] = task.get("taskId")
        return task

    def stop(self, id: int):
        """stop"""
        response = self._api.post("tasks.stop", {ApiField.ID: id})
        return self.Status(response.json()[ApiField.STATUS])

    def get_import_files_list(self, id: int) -> Dict or None:
        """get_import_files_list"""
        response = self._api.post("tasks.import.files_list", {ApiField.ID: id})
        return response.json() if (response is not None) else None

    def download_import_file(self, id, file_path, save_path):
        """download_import_file"""
        response = self._api.post(
            "tasks.import.download_file",
            {ApiField.ID: id, ApiField.FILENAME: file_path},
            stream=True,
        )

        ensure_base_path(save_path)
        with open(save_path, "wb") as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)

    def create_task_detached(self, workspace_id: int, task_type: Optional[str] = None):
        """create_task_detached"""
        response = self._api.post(
            "tasks.run.python",
            {
                ApiField.WORKSPACE_ID: workspace_id,
                ApiField.SCRIPT: "xxx",
                ApiField.ADVANCED: {ApiField.IGNORE_AGENT: True},
            },
        )
        return response.json()[ApiField.TASK_ID]

    def submit_logs(self, logs) -> None:
        """submit_logs"""
        response = self._api.post("tasks.logs.add", {ApiField.LOGS: logs})
        # return response.json()[ApiField.TASK_ID]

    def upload_files(
        self,
        task_id: int,
        abs_paths: List[str],
        names: List[str],
        progress_cb: Optional[Callable] = None,
    ) -> None:
        """upload_files"""
        if len(abs_paths) != len(names):
            raise RuntimeError("Inconsistency: len(abs_paths) != len(names)")

        hashes = []
        if len(abs_paths) == 0:
            return

        hash_to_items = defaultdict(list)
        hash_to_name = defaultdict(list)
        for idx, item in enumerate(zip(abs_paths, names)):
            path, name = item
            item_hash = get_file_hash(path)
            hashes.append(item_hash)
            hash_to_items[item_hash].append(path)
            hash_to_name[item_hash].append(name)

        unique_hashes = set(hashes)
        remote_hashes = self._api.image.check_existing_hashes(list(unique_hashes))
        new_hashes = unique_hashes - set(remote_hashes)

        # @TODO: upload remote hashes
        if len(remote_hashes) != 0:
            files = []
            for hash in remote_hashes:
                for name in hash_to_name[hash]:
                    files.append({ApiField.NAME: name, ApiField.HASH: hash})
            for batch in batched(files):
                resp = self._api.post(
                    "tasks.files.bulk.add-by-hash",
                    {ApiField.TASK_ID: task_id, ApiField.FILES: batch},
                )
        if progress_cb is not None:
            progress_cb(len(remote_hashes))

        for batch in batched(list(zip(abs_paths, names, hashes))):
            content_dict = OrderedDict()
            for idx, item in enumerate(batch):
                path, name, hash = item
                if hash in remote_hashes:
                    continue
                content_dict["{}".format(idx)] = json.dumps({"fullpath": name, "hash": hash})
                content_dict["{}-file".format(idx)] = (name, open(path, "rb"), "")

            if len(content_dict) > 0:
                encoder = MultipartEncoder(fields=content_dict)
                resp = self._api.post("tasks.files.bulk.upload", encoder)
                if progress_cb is not None:
                    progress_cb(len(content_dict))

    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2}
    # }
    # {
    #     obj: {new_val: 1}
    # }
    # // apped: true, recursive: false
    # {
    #     data: {my_val: 1}
    #     obj: {new_val: 1}
    # }(edited)
    # // append: false, recursive: false
    # {
    #     obj: {new_val: 1}
    # }(edited)
    #
    # 16: 32
    # // append: true, recursive: true
    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2, new_val: 1}
    # }

    def set_fields(self, task_id: int, fields: List) -> Dict:
        """set_fields"""
        for idx, obj in enumerate(fields):
            for key in [ApiField.FIELD, ApiField.PAYLOAD]:
                if key not in obj:
                    raise KeyError("Object #{} does not have field {!r}".format(idx, key))
        data = {ApiField.TASK_ID: task_id, ApiField.FIELDS: fields}
        resp = self._api.post("tasks.data.set", data)
        return resp.json()

    def set_fields_from_dict(self, task_id: int, d: Dict) -> Dict:
        """set_fields_from_dict"""
        fields = []
        for k, v in d.items():
            fields.append({ApiField.FIELD: k, ApiField.PAYLOAD: v})
        return self.set_fields(task_id, fields)

    def set_field(
        self,
        task_id: int,
        field: Dict,
        payload: Dict,
        append: Optional[bool] = False,
        recursive: Optional[bool] = False,
    ) -> Dict:
        """set_field"""
        fields = [
            {
                ApiField.FIELD: field,
                ApiField.PAYLOAD: payload,
                ApiField.APPEND: append,
                ApiField.RECURSIVE: recursive,
            }
        ]
        return self.set_fields(task_id, fields)

    def get_fields(self, task_id, fields: List):
        """get_fields"""
        data = {ApiField.TASK_ID: task_id, ApiField.FIELDS: fields}
        resp = self._api.post("tasks.data.get", data)
        return resp.json()["result"]

    def get_field(self, task_id: int, field: Dict):
        """get_field"""
        result = self.get_fields(task_id, [field])
        return result[field]

    def _validate_checkpoints_support(self, task_id):
        """_validate_checkpoints_support"""
        info = self.get_info_by_id(task_id)
        if info["type"] != str(TaskApi.PluginTaskType.TRAIN):
            raise RuntimeError(
                "Task (id={!r}) has type {!r}. "
                "Checkpoints are available only for tasks of type {!r}".format()
            )

    def list_checkpoints(self, task_id: int):
        """list_checkpoints"""
        self._validate_checkpoints_support(task_id)
        resp = self._api.post("tasks.checkpoints.list", {ApiField.ID: task_id})
        return resp.json()

    def delete_unused_checkpoints(self, task_id: int) -> Dict:
        """delete_unused_checkpoints"""
        self._validate_checkpoints_support(task_id)
        resp = self._api.post("tasks.checkpoints.clear", {ApiField.ID: task_id})
        return resp.json()

    def _set_output(self):
        """_set_output"""
        pass

    def set_output_project(
        self, task_id: int, project_id: int, project_name: Optional[str] = None
    ) -> Dict:
        """set_output_project"""
        if project_name is None:
            project = self._api.project.get_info_by_id(project_id, raise_error=True)
            project_name = project.name

        output = {ApiField.PROJECT: {ApiField.ID: project_id, ApiField.TITLE: project_name}}
        resp = self._api.post(
            "tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output}
        )
        return resp.json()

    def set_output_report(self, task_id: int, file_id: int, file_name: str) -> Dict:
        """set_output_report"""
        return self._set_custom_output(
            task_id, file_id, file_name, description="Report", icon="zmdi zmdi-receipt"
        )

    def _set_custom_output(
        self,
        task_id,
        file_id,
        file_name,
        file_url=None,
        description="File",
        icon="zmdi zmdi-file-text",
        color="#33c94c",
        background_color="#d9f7e4",
        download=False,
    ):
        """_set_custom_output"""
        if file_url is None:
            file_url = self._api.file.get_url(file_id)

        output = {
            ApiField.GENERAL: {
                "icon": {
                    "className": icon,
                    "color": color,
                    "backgroundColor": background_color,
                },
                "title": file_name,
                "titleUrl": file_url,
                "download": download,
                "description": description,
            }
        }
        resp = self._api.post(
            "tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output}
        )
        return resp.json()

    def set_output_archive(
        self, task_id: int, file_id: int, file_name: str, file_url: Optional[str] = None
    ) -> Dict:
        """set_output_archive"""
        if file_url is None:
            file_url = self._api.file.get_info_by_id(file_id).storage_path
        return self._set_custom_output(
            task_id,
            file_id,
            file_name,
            file_url=file_url,
            description="Download archive",
            icon="zmdi zmdi-archive",
            download=True,
        )

    def set_output_file_download(
        self,
        task_id: int,
        file_id: int,
        file_name: str,
        file_url: Optional[str] = None,
        download: Optional[bool] = True,
    ) -> Dict:
        """set_output_file_download"""
        if file_url is None:
            file_url = self._api.file.get_info_by_id(file_id).storage_path
        return self._set_custom_output(
            task_id,
            file_id,
            file_name,
            file_url=file_url,
            description="Download file",
            icon="zmdi zmdi-file",
            download=download,
        )

    def send_request(
        self,
        task_id: int,
        method: str,
        data: Dict,
        context: Optional[Dict] = {},
        skip_response: bool = False,
        timeout: Optional[int] = 60,
    ):
        """send_request"""
        if type(data) is not dict:
            raise TypeError("data argument has to be a dict")
        resp = self._api.post(
            "tasks.request.direct",
            {
                ApiField.TASK_ID: task_id,
                ApiField.COMMAND: method,
                ApiField.CONTEXT: context,
                ApiField.STATE: data,
                "skipResponse": skip_response,
                "timeout": timeout,
            },
        )
        return resp.json()

    def set_output_directory(self, task_id, file_id, directory_path):
        """set_output_directory"""
        return self._set_custom_output(
            task_id,
            file_id,
            directory_path,
            description="Directory",
            icon="zmdi zmdi-folder",
        )

    def update_meta(
        self, id: int, data: dict, agent_storage_folder: str = None, relative_app_dir: str = None
    ):
        """
        Update given task metadata
        :param id: int — task id
        :param data: dict — meta data to update
        """
        if type(data) == dict:
            data.update({"id": id})
            if agent_storage_folder is None and relative_app_dir is not None:
                raise ValueError(
                    "Both arguments (agent_storage_folder and relative_app_dir) has to be defined or None"
                )
            if agent_storage_folder is not None and relative_app_dir is None:
                raise ValueError(
                    "Both arguments (agent_storage_folder and relative_app_dir) has to be defined or None"
                )
            if agent_storage_folder is not None and relative_app_dir is not None:
                data["agentStorageFolder"] = {
                    "hostDir": agent_storage_folder,
                    "folder": relative_app_dir,
                }

        self._api.post("tasks.meta.update", data)



# coding: utf-8
import math


VAL_EVERY = 'val_every'


class EvalPlanner:
    """
    Helps determine whether to start validation in current train step or not.

    Args:
        epochs: Number of train epochs.
        val_every:  Validation period by epoch (value 0.5 mean 2 validations per epoch).
    """
    def __init__(self, epochs, val_every):
        self.epochs = epochs
        self.val_every = val_every
        self.total_val_cnt = self.validations_cnt(epochs, val_every)
        self._val_cnt = 0

    @property
    def performed_val_cnt(self):
        """int: Number of performed validations."""
        return self._val_cnt

    @staticmethod
    def validations_cnt(ep_float, val_every):
        res = math.floor(ep_float / val_every + 1e-9)
        return res

    def validation_performed(self):
        """Increments number of performed validations"""
        self._val_cnt += 1

    def need_validation(self, epoch_flt):
        """
        Determines whether to start validation or not.

        Args:
            epoch_flt: Current train step.
        Returns:
            True if validation is needed in current ste, False otherwise.
        """
        req_val_cnt = self.validations_cnt(epoch_flt, self.val_every)
        need_val = req_val_cnt > self._val_cnt
        return need_val



# coding: utf-8
"""download/upload/manipulate files from/to Supervisely team files"""

from __future__ import annotations

from typing import NamedTuple, List, Dict, Optional, Callable

import os
import shutil
import tarfile
from pathlib import Path
import urllib
import re

from supervisely._utils import batched, rand_str
from supervisely.api.module_api import ModuleApiBase, ApiField
import supervisely.io.fs as sly_fs
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
import mimetypes
from supervisely.imaging.image import write_bytes, get_hash
from supervisely.task.progress import Progress
from supervisely.io.fs_cache import FileCache
from supervisely.io.fs import (
    get_file_hash,
    get_file_name,
    get_file_ext,
    get_file_size,
    list_files_recursively,
    silent_remove,
    ensure_base_path,
    get_file_name_with_ext,
)
from supervisely.sly_logger import logger
import supervisely.io.env as env


class FileInfo(NamedTuple):
    """ """

    team_id: int
    id: int
    user_id: int
    name: str
    hash: str
    path: str
    storage_path: str
    mime: str
    ext: str
    sizeb: int
    created_at: str
    updated_at: str
    full_storage_url: str


class FileApi(ModuleApiBase):
    """
    API for working with Files. :class:`FileApi<FileApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        team_id = 8
        file_path = "/999_App_Test/"
        files = api.file.list(team_id, file_path)
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple FileInfo information about File.

        :Example:

         .. code-block:: python

            FileInfo(team_id=8,
                     id=7660,
                     user_id=7,
                     name='00135.json',
                     hash='z7Hv9a7WIC5HIJrfX/69KVrvtDaLqucSprWHoCxyq0M=',
                     path='/999_App_Test/ds1/00135.json',
                     storage_path='/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/y/P/rn/...json',
                     mime='application/json',
                     ext='json',
                     sizeb=261,
                     created_at='2021-01-11T09:04:17.959Z',
                     updated_at='2021-01-11T09:04:17.959Z',
                     full_storage_url='http://supervise.ly/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/y/P/rn/...json')
        """
        return [
            ApiField.TEAM_ID,
            ApiField.ID,
            ApiField.USER_ID,
            ApiField.NAME,
            ApiField.HASH,
            ApiField.PATH,
            ApiField.STORAGE_PATH,
            ApiField.MIME2,
            ApiField.EXT2,
            ApiField.SIZEB2,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.FULL_STORAGE_URL,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **FileInfo**.
        """
        return "FileInfo"

    def list_on_agent(self, team_id: int, path: str, recursive: bool = True) -> List[Dict]:
        if self.is_on_agent(path) is False:
            raise ValueError(f"Data is not on agent: {path}")

        agent_id, path_in_agent_folder = self.parse_agent_id_and_path(path)
        dirs_queue: List[str] = [path_in_agent_folder]

        results = []
        while len(dirs_queue) > 0:
            cur_dir = dirs_queue.pop(0)
            if cur_dir.endswith("/") is False:
                cur_dir += "/"
            response = self._api.post(
                "agents.storage.list",
                {ApiField.ID: agent_id, ApiField.TEAM_ID: team_id, ApiField.PATH: cur_dir},
            )
            items = response.json()
            for item in items:
                if item["type"] == "file":
                    results.append(item)
                elif item["type"] == "directory" and recursive is True:
                    dirs_queue.append(os.path.join(cur_dir, item["name"]))

        return results

    def list(self, team_id: int, path: str, recursive: bool = True) -> List[Dict]:
        """
        List of files in the Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param path: Path to File or Directory.
        :type path: str
        :param recursive: If True return all files recursively.
        :type recursive: bool
        :return: List of all Files with information. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[dict]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            team_id = 8
            file_path = "/999_App_Test/"
            files = api.file.list(team_id, file_path)

            print(files)
            # Output: [
            #     {
            #         "id":7660,
            #         "userId":7,
            #         "path":"/999_App_Test/ds1/00135.json",
            #         "storagePath":"/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/y/P/rn/...json",
            #         "meta":{
            #             "ext":"json",
            #             "mime":"application/json",
            #             "size":261
            #         },
            #         "createdAt":"2021-01-11T09:04:17.959Z",
            #         "updatedAt":"2021-01-11T09:04:17.959Z",
            #         "hash":"z7Wv1a7WIC5HIJrfX/69XXrqtDaLxucSprWHoCxyq0M=",
            #         "fullStorageUrl":"http://supervise.ly/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/y/P/rn/...json",
            #         "teamId":8,
            #         "name":"00135.json"
            #     },
            #     {
            #         "id":7661,
            #         "userId":7,
            #         "path":"/999_App_Test/ds1/01587.json",
            #         "storagePath":"/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/9/k/Hs/...json",
            #         "meta":{
            #             "ext":"json",
            #             "mime":"application/json",
            #             "size":252
            #         },
            #         "createdAt":"2021-01-11T09:04:18.099Z",
            #         "updatedAt":"2021-01-11T09:04:18.099Z",
            #         "hash":"La9+XtF2+cTlAqUE/I72e/xS12LqyH1+z<3T+SgD4CTU=",
            #         "fullStorageUrl":"http://supervise.ly/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/9/k/Hs/...json",
            #         "teamId":8,
            #         "name":"01587.json"
            #     }
            # ]
        """

        if not path.endswith(os.path.sep) and recursive is False:
            path += os.path.sep
        if self.is_on_agent(path) is True:
            return self.list_on_agent(team_id, path, recursive)

        response = self._api.post(
            "file-storage.list",
            {ApiField.TEAM_ID: team_id, ApiField.PATH: path, ApiField.RECURSIVE: recursive},
        )
        return response.json()

    def list2(self, team_id: int, path: str, recursive: bool = True) -> List[FileInfo]:
        """
        List of files in the Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param path: Path to File or Directory.
        :type path: str
        :param recursive: If True return all FileInfos recursively.
        :type recursive: bool
        :return: List of all Files with information. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[FileInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            team_id = 9
            file_path = "/My_App_Test/"
            files = api.file.list2(team_id, file_path)

            print(files)
            # Output: [
            # FileInfo(team_id=9, id=18421, user_id=8, name='5071_3734_mot_video_002.tar.gz', hash='+0nrNoDjBxxJA...
            # FileInfo(team_id=9, id=18456, user_id=8, name='5164_4218_mot_video_bitmap.tar.gz', hash='fwtVI+iptY...
            # FileInfo(team_id=9, id=18453, user_id=8, name='all_vars.tar', hash='TVkUE+K1bnEb9QrdEm9akmHm/QEWPJK...
            # ]
        """
        items = self.list(team_id=team_id, path=path, recursive=recursive)
        results = [self._convert_json_info(info_json) for info_json in items]
        return results

    def listdir(self, team_id: int, path: str, recursive: bool = False) -> List[str]:
        """
        List dirs and files in the `path` dir.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param path: Path to directory.
        :type path: str
        :param recursive: If True return all paths recursively.
        :type recursive: bool
        :return: List of paths
        :rtype: :class:`List[str]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            team_id = 8
            path = "/999_App_Test/"
            files = api.file.listdir(team_id, path)

            print(files)
            # Output: ["/999_App_Test/ds1", "/999_App_Test/image.png"]
        """
        files = self.list(team_id, path, recursive)
        files_paths = [file["path"] for file in files]
        return files_paths

    def get_directory_size(self, team_id: int, path: str) -> int:
        """
        Get directory size in the Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param path: Path to Directory.
        :type path: str
        :return: Directory size in the Team Files
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            team_id = 9
            path = "/My_App_Test/"
            size = api.file.get_directory_size(team_id, path)

            print(size)
            # Output: 3478687
        """
        dir_size = 0
        file_infos = self.list2(team_id, path)
        for file_info in file_infos:
            dir_size += file_info.sizeb
        return dir_size

    def _download(
        self, team_id, remote_path, local_save_path, progress_cb=None
    ):  # TODO: progress bar
        response = self._api.post(
            "file-storage.download",
            {ApiField.TEAM_ID: team_id, ApiField.PATH: remote_path},
            stream=True,
        )
        # print(response.headers)
        # print(response.headers['Content-Length'])
        ensure_base_path(local_save_path)
        with open(local_save_path, "wb") as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)
                if progress_cb is not None:
                    progress_cb(len(chunk))

    def download(
        self,
        team_id: int,
        remote_path: str,
        local_save_path: str,
        cache: Optional[FileCache] = None,
        progress_cb: Optional[Callable] = None,
    ) -> None:
        """
        Download File from Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param remote_path: Path to File in Team Files.
        :type remote_path: str
        :param local_save_path: Local save path.
        :type local_save_path: str
        :param cache: optional
        :type cache: FileCache, optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: Progress, optional
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            path_to_file = "/999_App_Test/ds1/01587.json"
            local_save_path = "/home/admin/Downloads/01587.json"

            api.file.download(8, path_to_file, local_save_path)
        """
        if self.is_on_agent(remote_path):
            self.download_from_agent(remote_path, local_save_path, progress_cb)
            return

        if cache is None:
            self._download(team_id, remote_path, local_save_path, progress_cb)
        else:
            file_info = self.get_info_by_path(team_id, remote_path)
            if file_info.hash is None:
                self._download(team_id, remote_path, local_save_path, progress_cb)
            else:
                cache_path = cache.check_storage_object(file_info.hash, get_file_ext(remote_path))
                if cache_path is None:
                    # file not in cache
                    self._download(team_id, remote_path, local_save_path, progress_cb)
                    if file_info.hash != get_file_hash(local_save_path):
                        raise KeyError(
                            f"Remote and local hashes are different (team id: {team_id}, file: {remote_path})"
                        )
                    cache.write_object(local_save_path, file_info.hash)
                else:
                    cache.read_object(file_info.hash, local_save_path)
                    if progress_cb is not None:
                        progress_cb(get_file_size(local_save_path))

    def is_on_agent(self, remote_path: str):
        return sly_fs.is_on_agent(remote_path)

    def parse_agent_id_and_path(self, remote_path: str) -> int:
        return sly_fs.parse_agent_id_and_path(remote_path)

    def download_from_agent(
        self,
        remote_path: str,
        local_save_path: str,
        progress_cb: Optional[Callable] = None,
    ) -> None:
        agent_id, path_in_agent_folder = self.parse_agent_id_and_path(remote_path)
        if (
            agent_id == env.agent_id(raise_not_found=False)
            and env.agent_storage(raise_not_found=False) is not None
        ):
            path_on_agent = os.path.normpath(env.agent_storage() + path_in_agent_folder)
            logger.info(f"Optimized download from agent: {path_on_agent}")
            sly_fs.copy_file(path_on_agent, local_save_path)
            if progress_cb is not None:
                progress_cb(sly_fs.get_file_size(path_on_agent))
            return

        response = self._api.post(
            "agents.storage.download",
            {ApiField.ID: agent_id, ApiField.PATH: path_in_agent_folder},
            stream=True,
        )
        ensure_base_path(local_save_path)
        with open(local_save_path, "wb") as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)
                if progress_cb is not None:
                    progress_cb(len(chunk))

    def download_directory(
        self,
        team_id: int,
        remote_path: str,
        local_save_path: str,
        progress_cb: Optional[Callable] = None,
    ) -> None:
        """
        Download Directory from Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param remote_path: Path to Directory in Team Files.
        :type remote_path: str
        :param local_save_path: Local save path.
        :type local_save_path: str
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: Progress, optional
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            path_to_dir = "/My_App_Test/ds1"
            local_save_path = "/home/admin/Downloads/My_local_test"

            api.file.download_directory(9, path_to_dir, local_save_path)
        """
        if not remote_path.endswith(os.path.sep):
            remote_path += os.path.sep

        if self.is_on_agent(remote_path) is True:
            agent_id, path_in_agent_folder = self.parse_agent_id_and_path(remote_path)
            if (
                agent_id == env.agent_id(raise_not_found=False)
                and env.agent_storage(raise_not_found=False) is not None
            ):
                dir_on_agent = os.path.normpath(env.agent_storage() + path_in_agent_folder)
                logger.info(f"Optimized download from agent: {dir_on_agent}")
                sly_fs.copy_dir_recursively(dir_on_agent, local_save_path)
                return

        local_temp_archive = os.path.join(local_save_path, "temp.tar")
        self.download(team_id, remote_path, local_temp_archive, cache=None, progress_cb=progress_cb)
        tr = tarfile.open(local_temp_archive)
        tr.extractall(local_save_path)
        silent_remove(local_temp_archive)
        temp_dir = os.path.join(local_save_path, rand_str(10))
        to_move_dir = os.path.join(local_save_path, os.path.basename(os.path.normpath(remote_path)))
        os.rename(to_move_dir, temp_dir)
        file_names = os.listdir(temp_dir)
        for file_name in file_names:
            shutil.move(os.path.join(temp_dir, file_name), local_save_path)
        shutil.rmtree(temp_dir)

    def _upload_legacy(self, team_id, src, dst):
        """ """

        def path_to_bytes_stream(path):
            return open(path, "rb")

        item = get_file_name_with_ext(dst)
        content_dict = {}
        content_dict[ApiField.NAME] = item

        dst_dir = os.path.dirname(dst)
        if not dst_dir.endswith(os.path.sep):
            dst_dir += os.path.sep
        content_dict[ApiField.PATH] = dst_dir  # os.path.basedir ...
        content_dict["file"] = (
            item,
            path_to_bytes_stream(src),
            mimetypes.MimeTypes().guess_type(src)[0],
        )
        encoder = MultipartEncoder(fields=content_dict)
        resp = self._api.post("file-storage.upload?teamId={}".format(team_id), encoder)
        return resp.json()

    def upload(
        self, team_id: int, src: str, dst: str, progress_cb: Optional[Callable] = None
    ) -> FileInfo:
        """
        Upload File to Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param src: Local source file path.
        :type src: str
        :param dst: Path to File in Team Files.
        :type dst: str
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: Progress, optional
        :return: Information about File. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`FileInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_path = "/home/admin/Downloads/01587.json"
            dst_remote_path = "/999_App_Test/ds1/01587.json"

            api.file.upload(8, src_path, dst_remote_path)
        """
        return self.upload_bulk(team_id, [src], [dst], progress_cb)[0]

    def upload_bulk(
        self,
        team_id: int,
        src_paths: List[str],
        dst_paths: List[str],
        progress_cb: Optional[Callable] = None,
    ) -> List[FileInfo]:
        """
        Upload Files to Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param src: Local source file paths.
        :type src: List[str]
        :param dst: Destination paths for Files to Team Files.
        :type dst: List[str]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: Progress, optional
        :return: Information about Files. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[FileInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_paths = ["/home/admin/Downloads/01587.json", "/home/admin/Downloads/01588.json","/home/admin/Downloads/01589.json"]
            dst_remote_paths = ["/999_App_Test/ds1/01587.json", "/999_App_Test/ds1/01588.json", "/999_App_Test/ds1/01589.json"]

            api.file.upload_bulk(8, src_paths, dst_remote_paths)
        """

        def path_to_bytes_stream(path):
            return open(path, "rb")

        content_dict = []
        for idx, (src, dst) in enumerate(zip(src_paths, dst_paths)):
            name = get_file_name_with_ext(dst)
            content_dict.append((ApiField.NAME, name))
            dst_dir = os.path.dirname(dst)
            if not dst_dir.endswith("/"):
                dst_dir += "/"
            content_dict.append((ApiField.PATH, dst_dir))
            content_dict.append(
                (
                    "file",
                    (
                        name,
                        path_to_bytes_stream(src),
                        mimetypes.MimeTypes().guess_type(src)[0],
                    ),
                )
            )
        encoder = MultipartEncoder(fields=content_dict)

        # progress = None
        # if progress_logger is not None:
        #     progress = Progress("Uploading", encoder.len, progress_logger, is_size=True)
        #
        # def _print_progress(monitor):
        #     if progress is not None:
        #         progress.set_current_value(monitor.bytes_read)
        #         print(monitor.bytes_read, monitor.len)
        # last_percent = 0
        # def _progress_callback(monitor):
        #     cur_percent = int(monitor.bytes_read * 100.0 / monitor.len)
        #     if cur_percent - last_percent > 10 or cur_percent == 100:
        #         api.task.set_fields(task_id, [{"field": "data.previewProgress", "payload": cur_percent}])
        #     last_percent = cur_percent

        if progress_cb is None:
            data = encoder
        else:
            data = MultipartEncoderMonitor(encoder, progress_cb)
        resp = self._api.post("file-storage.bulk.upload?teamId={}".format(team_id), data)
        results = [self._convert_json_info(info_json) for info_json in resp.json()]
        return results

    def rename(self, old_name: str, new_name: str) -> None:
        """
        Renames file in Team Files

        :param old_name: Old File name.
        :type old_name: str
        :param new_name: New File name.
        :type new_name: str
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            # NotImplementedError('Method is not supported')
        """
        pass

    def remove_from_agent(self, team_id: int, path: str) -> None:
        raise NotImplementedError()
        agent_id, path_in_agent_folder = self.parse_agent_id_and_path(path)
        if (
            agent_id == env.agent_id(raise_not_found=False)
            and env.agent_storage(raise_not_found=False) is not None
        ):
            # path_on_agent = os.path.normpath(env.agent_storage() + path_in_agent_folder)
            # logger.info(f"Optimized download from agent: {path_on_agent}")
            # sly_fs.copy_file(path_on_agent, local_save_path)
            return

    def remove(self, team_id: int, path: str) -> None:
        """
        Removes a file from the Team Files. If the specified path is a directory,
        the entire directory (including all recursively included files) will be removed.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param path: Path in Team Files.
        :type path: str
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.file.remove(8, "/999_App_Test/ds1/01587.json") # remove file
            api.file.remove(8, "/999_App_Test/ds1/") # remove folder
        """

        if self.is_on_agent(path) is True:
            # self.remove_from_agent(team_id, path)
            logger.warn(
                f"Data '{path}' is on agent. Method does not support agent storage. Remove your data manually on the computer with agent."
            )
            return

        resp = self._api.post(
            "file-storage.remove", {ApiField.TEAM_ID: team_id, ApiField.PATH: path}
        )

    def remove_file(self, team_id: int, path: str) -> None:
        """
        Removes file from Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param path: Path to File in Team Files.
        :type path: str
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.file.remove_file(8, "/999_App_Test/ds1/01587.json")
        """

        file_info = self.get_info_by_path(team_id, path)

        if file_info is None:
            raise ValueError(
                f"File not found in Team files. Maybe you entered directory? (Path: '{path}')"
            )

        self.remove(team_id, path)

    def remove_dir(self, team_id: int, path: str) -> None:
        """
        Removes folder from Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param path: Path to folder in Team Files.
        :type path: str
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.file.remove_dir(8, "/999_App_Test/ds1/")
        """

        if not path.endswith("/"):
            raise ValueError("Please add a slash in the end to recognize path as a directory.")

        if not self.dir_exists(team_id, path):
            raise ValueError(f"Folder not found in Team files. (Path: '{path}')")

        self.remove(team_id, path)

    def remove_batch(
        self,
        team_id: int,
        paths: List[str],
        progress_cb: Optional[Callable] = None,
    ) -> None:
        """
        Removes list of files from Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param paths: List of paths to Files in Team Files.
        :type paths: List[str]
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

        .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            paths_to_del = [
                "/999_App_Test/ds1/01587.json",
                "/999_App_Test/ds1/01588.json",
                "/999_App_Test/ds1/01587.json"
            ]
            api.file.remove(8, paths_to_del)
        """

        for paths_batch in batched(paths, batch_size=100):
            for path in paths_batch:
                if self.is_on_agent(path) is True:
                    logger.warn(
                        f"Data '{path}' is on agent. File skipped. Method does not support agent storage. Remove your data manually on the computer with agent."
                    )
                    paths_batch.remove(path)

            self._api.post(
                "file-storage.bulk.remove", {ApiField.TEAM_ID: team_id, ApiField.PATHS: paths_batch}
            )
            if progress_cb is not None:
                progress_cb(len(paths_batch))

    def exists(self, team_id: int, remote_path: str) -> bool:
        """
        Checks if file exists in Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param remote_path: Remote path to File in Team Files.
        :type remote_path: str
        :return: True if file exists, otherwise False
        :rtype: :class:`bool`
        :Usage example:

         .. code-block:: python

           import supervisely as sly

           os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
           os.environ['API_TOKEN'] = 'Your Supervisely API Token'
           api = sly.Api.from_env()

           file = api.file.exists(8, "/999_App_Test/ds1/02163.json") # True
           file = api.file.exists(8, "/999_App_Test/ds1/01587.json") # False
        """
        path_infos = self.list(team_id, remote_path)
        for info in path_infos:
            if info["path"] == remote_path:
                return True
        return False

    def dir_exists(self, team_id: int, remote_directory: str) -> bool:
        """
        Checks if directory exists in Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param remote_path: Remote path to directory in Team Files.
        :type remote_path: str
        :return: True if directory exists, otherwise False
        :rtype: :class:`bool`
        :Usage example:

         .. code-block:: python

           import supervisely as sly

           os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
           os.environ['API_TOKEN'] = 'Your Supervisely API Token'
           api = sly.Api.from_env()

           file = api.file.dir_exists(8, "/999_App_Test/")   # True
           file = api.file.dir_exists(8, "/10000_App_Test/") # False
        """
        files_infos = self.list(team_id, remote_directory)
        if len(files_infos) > 0:
            return True
        return False

    def get_free_name(self, team_id: int, path: str) -> str:
        """
        Adds suffix to the end of the file name.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param path: Remote path to file in Team Files.
        :type path: str
        :return: New File name with suffix at the end
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

           import supervisely as sly

           os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
           os.environ['API_TOKEN'] = 'Your Supervisely API Token'
           api = sly.Api.from_env()

           file = api.file.get_free_name(8, "/999_App_Test/ds1/02163.json")
           print(file)
           # Output: /999_App_Test/ds1/02163_000.json
        """
        directory = Path(path).parent
        name = get_file_name(path)
        ext = get_file_ext(path)
        res_name = name
        suffix = 0

        def _combine(suffix: int = None):
            res = "{}/{}".format(directory, res_name)
            if suffix is not None:
                res += "_{:03d}".format(suffix)
            if ext:
                res += "{}".format(ext)
            return res

        res_path = _combine()
        while self.exists(team_id, res_path):
            res_path = _combine(suffix)
            suffix += 1
        return res_path

    def get_url(self, file_id: int) -> str:
        """
        Gets URL for the File by ID.

        :param file_id: File ID in Supervisely.
        :type file_id: int
        :return: File URL
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

           import supervisely as sly

           os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
           os.environ['API_TOKEN'] = 'Your Supervisely API Token'
           api = sly.Api.from_env()

           file_id = 7660
           file_url = sly.api.file.get_url(file_id)
           print(file_url)
           # Output: http://supervise.ly/files/7660
        """
        return f"/files/{file_id}"

    def get_info_by_path(self, team_id: int, remote_path: str) -> FileInfo:
        """
        Gets File information by path in Team Files.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param remote_path: Remote path to file in Team Files.
        :type remote_path: str
        :return: Information about File. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`FileInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            file_path = "/999_App_Test/ds1/00135.json"
            file_info = api.file.get_info_by_id(8, file_path)
            print(file_info)
            # Output: FileInfo(team_id=8,
            #                  id=7660,
            #                  user_id=7,
            #                  name='00135.json',
            #                  hash='z7Hv9a7WIC5HIJrfX/69KVrvtDaLqucSprWHoCxyq0M=',
            #                  path='/999_App_Test/ds1/00135.json',
            #                  storage_path='/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/y/P/rn/...json',
            #                  mime='application/json',
            #                  ext='json',
            #                  sizeb=261,
            #                  created_at='2021-01-11T09:04:17.959Z',
            #                  updated_at='2021-01-11T09:04:17.959Z',
            #                  full_storage_url='http://supervise.ly/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/y/P/rn/...json')
        """
        if self.is_on_agent(remote_path) is True:
            path_infos = self.list_on_agent(team_id, os.path.dirname(remote_path), recursive=False)
            for info in path_infos:
                if info["path"] == remote_path:
                    return self._convert_json_info(info)
        else:
            path_infos = self.list(team_id, remote_path)
            for info in path_infos:
                if info["path"] == remote_path:
                    return self._convert_json_info(info)
        return None

    def _convert_json_info(self, info: dict, skip_missing=True) -> FileInfo:
        """ """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        # if res.storage_path is not None:
        #    res = res._replace(full_storage_url=urllib.parse.urljoin(self._api.server_address, res.storage_path))
        return FileInfo(**res._asdict())

    def get_info_by_id(self, id: int) -> FileInfo:
        """
        Gets information about File by ID.

        :param id: File ID in Supervisely.
        :type id: int
        :return: Information about File. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`FileInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            file_id = 7660
            file_info = api.file.get_info_by_id(file_id)
            print(file_info)
            # Output: FileInfo(team_id=8,
            #                  id=7660,
            #                  user_id=7,
            #                  name='00135.json',
            #                  hash='z7Hv9a7WIC5HIJrfX/69KVrvtDaLqucSprWHoCxyq0M=',
            #                  path='/999_App_Test/ds1/00135.json',
            #                  storage_path='/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/y/P/rn/...json',
            #                  mime='application/json',
            #                  ext='json',
            #                  sizeb=261,
            #                  created_at='2021-01-11T09:04:17.959Z',
            #                  updated_at='2021-01-11T09:04:17.959Z',
            #                  full_storage_url='http://supervise.ly/h5un6l2bnaz1vj8a9qgms4-public/teams_storage/8/y/P/rn/...json')
        """
        resp = self._api.post("file-storage.info", {ApiField.ID: id})
        return self._convert_json_info(resp.json())

    def get_free_dir_name(self, team_id: int, dir_path: str) -> str:
        """
        Adds suffix to the end of the Directory name.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param dir_path: Path to Directory in Team Files.
        :type dir_path: str
        :return: New Directory name with suffix at the end
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

           import supervisely as sly

           os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
           os.environ['API_TOKEN'] = 'Your Supervisely API Token'
           api = sly.Api.from_env()

           new_dir_name = api.file.get_free_dir_name(9, "/My_App_Test")
           print(new_dir_name)
           # Output: /My_App_Test_001
        """
        res_dir = dir_path.rstrip("/")
        suffix = 1
        while self.dir_exists(team_id, res_dir):
            res_dir = dir_path.rstrip("/") + f"_{suffix:03d}"
            suffix += 1
        return res_dir

    def upload_directory(
        self,
        team_id: int,
        local_dir: str,
        remote_dir: str,
        change_name_if_conflict: Optional[bool] = True,
        progress_size_cb: Optional[Callable] = None,
    ) -> str:
        """
        Upload Directory to Team Files from local path.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param local_dir: Path to local Directory.
        :type local_dir: str
        :param remote_dir: Path to Directory in Team Files.
        :type remote_dir: str
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :param progress_size_cb: Function for tracking download progress.
        :type progress_size_cb: Progress, optional
        :return: Path to Directory in Team Files
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            path_to_dir = "/My_App_Test/ds1"
            local_path = "/home/admin/Downloads/My_local_test"

            api.file.upload_directory(9, local_path, path_to_dir)
        """
        if self.dir_exists(team_id, remote_dir):
            if change_name_if_conflict is True:
                res_remote_dir = self.get_free_dir_name(team_id, remote_dir)
            else:
                raise FileExistsError(
                    f"Directory {remote_dir} already exists in your team (id={team_id})"
                )
        else:
            res_remote_dir = remote_dir

        local_files = list_files_recursively(local_dir)
        remote_files = [
            file.replace(local_dir.rstrip("/"), res_remote_dir.rstrip("/")) for file in local_files
        ]

        for local_paths_batch, remote_files_batch in zip(
            batched(local_files, batch_size=50), batched(remote_files, batch_size=50)
        ):

            self.upload_bulk(team_id, local_paths_batch, remote_files_batch, progress_size_cb)
        return res_remote_dir



# coding: utf-8

import os
import traceback
import logging
from supervisely.sly_logger import logger, EventType


SLY_DEBUG = "SLY_DEBUG"


def main_wrapper(main_name, main_func, *args, **kwargs):
    try:
        logger.debug("Main started.", extra={"main_name": main_name})
        main_func(*args, **kwargs)
    except Exception as e:
        logger.critical(
            "Unexpected exception in main.",
            exc_info=True,
            extra={
                "main_name": main_name,
                "event_type": EventType.TASK_CRASHED,
                "exc_str": str(e),
            },
        )
        logger.debug("Main finished: BAD.", extra={"main_name": main_name})

        if os.environ.get(SLY_DEBUG) or logging.getLevelName(logger.level) in ["TRACE", "DEBUG"]:
            raise
        else:
            os._exit(1)
    else:
        logger.debug("Main finished: OK.", extra={"main_name": main_name})


def function_wrapper(f, *args, **kwargs):
    try:
        return f(*args, **kwargs)
    except Exception as e:
        logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
        raise e


def catch_silently(fn, *args, **kwargs):
    try:
        return fn(*args, **kwargs)
    except Exception as e:
        logger.debug(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
        return None


def function_wrapper_nofail(f, *args, **kwargs):
    try:
        return f(*args, **kwargs)
    except Exception as e:
        logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})


def function_wrapper_external_logger(f, ext_logger, *args, **kwargs):
    try:
        return f(*args, **kwargs)
    except Exception as e:
        ext_logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
        raise e



# coding: utf-8

from copy import deepcopy
from supervisely._utils import take_with_default, validate_img_size
from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.volume_annotation import constants
from supervisely.volume_annotation.slice import Slice
from supervisely.sly_logger import logger

# example:
# "volumeMeta": {
#     "ACS": "RAS",
#     "intensity": { "max": 3071, "min": -3024 },
#     "windowWidth": 6095,
#     "rescaleSlope": 1,
#     "windowCenter": 23.5,
#     "channelsCount": 1,
#     "dimensionsIJK": { "x": 512, "y": 512, "z": 139 },
#     "IJK2WorldMatrix": [
#         0.7617189884185793, 0, 0, -194.238403081894, 0, 0.7617189884185793, 0,
#         -217.5384061336518, 0, 0, 2.5, -347.7500000000001, 0, 0, 0, 1
#     ],
#     "rescaleIntercept": 0
# },


class Plane(FrameCollection):
    item_type = Slice

    SAGITTAL = "sagittal"
    CORONAL = "coronal"
    AXIAL = "axial"
    _valid_names = [SAGITTAL, CORONAL, AXIAL, None]  # None for spatial figures

    @staticmethod
    def validate_name(name):
        if name not in Plane._valid_names:
            raise ValueError(
                f"Unknown plane {name}, valid names are {Plane._valid_names}"
            )

    def __init__(
        self, plane_name, img_size=None, slices_count=None, items=None, volume_meta=None
    ):
        Plane.validate_name(plane_name)
        self._name = plane_name

        if img_size is None and volume_meta is None:
            raise ValueError(
                "Both img_size and volume_meta are None, only one of them is allowed to be a None"
            )
        if slices_count is None and volume_meta is None:
            raise ValueError(
                "Both slices_count and volume_meta are None, only one of them is allowed to be a None"
            )
        self._img_size = take_with_default(
            img_size, Plane.get_img_size(self._name, volume_meta)
        )
        self._img_size = validate_img_size(self._img_size)

        self._slices_count = take_with_default(
            slices_count, Plane.get_slices_count(self._name, volume_meta)
        )

        super().__init__(items=items)

    @property
    def name(self):
        return self._name

    @property
    def slices_count(self):
        return self._slices_count

    @property
    def img_size(self):
        return deepcopy(self._img_size)

    @property
    def normal(self):
        return Plane.get_normal(self.name)

    def __str__(self):
        return super().__str__().replace("Frames", "Slices")

    @staticmethod
    def get_normal(name):
        Plane.validate_name(name)
        if name == Plane.SAGITTAL:
            return {"x": 1, "y": 0, "z": 0}
        if name == Plane.CORONAL:
            return {"x": 0, "y": 1, "z": 0}
        if name == Plane.AXIAL:
            return {"x": 0, "y": 0, "z": 1}

    @staticmethod
    def get_name(normal: dict):
        if normal == {"x": 1, "y": 0, "z": 0}:
            return Plane.SAGITTAL
        if normal == {"x": 0, "y": 1, "z": 0}:
            return Plane.CORONAL
        if normal == {"x": 0, "y": 0, "z": 1}:
            return Plane.AXIAL

    @staticmethod
    def get_img_size(name, volume_meta):
        Plane.validate_name(name)
        dimentions = volume_meta["dimensionsIJK"]
        # (height, width)
        height = None
        width = None
        if name == Plane.SAGITTAL:
            width = dimentions["y"]
            height = dimentions["z"]
        elif name == Plane.CORONAL:
            width = dimentions["x"]
            height = dimentions["z"]
        elif name == Plane.AXIAL:
            width = dimentions["x"]
            height = dimentions["y"]
        return [height, width]

    @staticmethod
    def get_slices_count(name, volume_meta):
        Plane.validate_name(name)
        dimentions = volume_meta["dimensionsIJK"]
        if name == Plane.SAGITTAL:
            return dimentions["x"]
        elif name == Plane.CORONAL:
            return dimentions["y"]
        elif name == Plane.AXIAL:
            return dimentions["z"]

    @classmethod
    def from_json(
        cls,
        data,
        plane_name,
        objects,
        img_size=None,
        slices_count=None,
        volume_meta=None,
        key_id_map=None,
    ):
        Plane.validate_name(plane_name)
        if plane_name != data[constants.NAME]:
            raise ValueError(
                f"Plane name {plane_name} differs from the name in json data {data[constants.NAME]}"
            )

        if Plane.get_normal(plane_name) != data[constants.NORMAL]:
            raise ValueError(
                f"Normal json data {data[constants.NORMAL]} is not valid for {plane_name}. It should be {Plane.get_normal(plane_name)}"
            )
        if slices_count is None:
            if volume_meta is None:
                raise ValueError(
                    "Both slices_count and volume_meta are None, only one of them is allowed to be a None"
                )
            else:
                slices_count = Plane.get_slices_count(plane_name, volume_meta)

        slices = []
        for slice_json in data[constants.SLICES]:
            slices.append(
                cls.item_type.from_json(
                    slice_json, objects, plane_name, slices_count, key_id_map
                )
            )

        return cls(plane_name, img_size, slices_count, slices, volume_meta)

    def to_json(self, key_id_map=None):
        json_slices = []
        for slice in self:
            slice: Slice
            json_slices.append(slice.to_json(key_id_map))

        return {
            constants.NAME: self.name,
            constants.NORMAL: self.normal,
            constants.SLICES: json_slices,
        }

    def validate_figures_bounds(self):
        for slice in self:
            slice: Slice
            slice.validate_figures_bounds(self.img_size)



from __future__ import annotations
from typing import Optional, Dict, Union
import uuid

from supervisely._utils import take_with_default
from supervisely.annotation.tag import Tag, TagJsonFields
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.pointcloud_annotation.constants import KEY, ID

class PointcloudTag(Tag):
    """
    PointcloudTag object for :class:`PointcloudAnnotation<supervisely.pointcloud_annotation.pointcloud_annotation.PointcloudAnnotation>`. :class:`PointcloudTag<PointcloudTag>` object is immutable.

    :param meta: General information about Pointcloud Tag.
    :type meta: :class:`TagMeta<supervisely.annotation.tag_meta.TagMeta>`
    :param value: Pointcloud Tag value. Depends on :class:`TagValueType<supervisely.annotation.tag_meta.TagValueType>` of :class:`TagMeta<supervisely.annotation.tag_meta.TagMeta>`.
    :type value: :class:`str` or :class:`int` or :class:`float` or :class:`NoneType`, optional
    :param key: uuid.UUID object.
    :type key: uuid.UUID, optional
    :param sly_id: Video Tag ID in Supervisely.
    :type sly_id: :class:`int`, optional
    :param labeler_login: Login of user who created PointcloudTag.
    :type labeler_login: :class:`str`, optional
    :param updated_at: Date and Time when PointcloudTag was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: :class:`str`, optional
    :param created_at: Date and Time when PointcloudTag was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: :class:`str`, optional
    :raises: :class:`ValueError`, If PointcloudTag value is incompatible to :class:`TagMeta<supervisely.annotation.tag_meta.TagMeta>` value type.

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
        # Now we can create a PointcloudTag using our TagMeta
        tag_dog = sly.PointcloudTag(meta_dog)
        # When you are creating a new Tag
        # Tag.value is automatically cross-checked against your TagMeta value type to make sure the value is valid.
        # If we now try to add a value to our newly created Tag, we receive "ValueError", because our TagMeta value type is "NONE"
        tag_dog = sly.PointcloudTag(meta_dog, value="Husky")
        # Output: ValueError: Tag dog can not have value Husky

        # Let's create another Tag with a string value type
        meta_cat = sly.TagMeta('cat', sly.TagValueType.ANY_STRING)
        tag_cat = sly.PointcloudTag(meta_cat, value="Fluffy")

        # Now let's create a Tag using TagMeta with "ONEOF_STRING" value type
        # In order to use "oneof_string value type", you must initialize a variable with possible values(see class TagMeta for more information)
        colors = ["brown", "white", "black", "red", "chocolate", "gold", "grey"]
        meta_coat_color = sly.TagMeta('coat color', sly.TagValueType.ONEOF_STRING, possible_values=colors)
        tag_coat_color = sly.PointcloudTag(meta_coat_color, value="white")

        # If given value is not in a list of possible Tags, ValueError will be raised
        tag_coat_color = sly.PointcloudTag(meta_coat_color, value="yellow")
        # Output: ValueError: Tag coat color can not have value yellow
    """
    def __init__(self, meta: TagMeta, value: Optional[Union[str, int, float]]=None,
                 key: Optional[uuid.UUID]=None, sly_id: Optional[int]=None, labeler_login: Optional[str]=None,
                 updated_at: Optional[str]=None, created_at: Optional[str]=None):
        super(PointcloudTag, self).__init__(meta, value=value, sly_id=sly_id, 
            labeler_login=labeler_login, 
            updated_at=updated_at, 
            created_at=created_at,
        )

        self._key = take_with_default(key, uuid.uuid4())

    def get_compact_str(self) -> str:
        """
        Get string with information about PointcloudTag name and value.

        :return: Information about PointcloudTag object
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            meta_cat = sly.TagMeta('cat', sly.TagValueType.ANY_STRING)
            tag_cat = sly.PointcloudTag(meta_cat, value="Fluffy")
            compact_tag_cat = tag_cat.get_compact_str()
            print(compact_tag_cat) # cat:Fluffy
        """
        return super(PointcloudTag, self).get_compact_str()

    def __eq__(self, other: PointcloudTag) -> bool:
        """
        Checks that 2 Pointcloud Tags are equal by comparing their meta and value.

        :param other: PointcloudTag object.
        :type other: :class:`PointcloudTag<PointcloudTag>`
        :return: True if comparable objects are equal, otherwise False
        :rtype: :class:`bool`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Let's create 2 identical Tags
            meta_lemon_1 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            tag_lemon_1 = sly.PointcloudTag(meta_lemon_1)

            meta_lemon_2 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            tag_lemon_2 = sly.PointcloudTag(meta_lemon_2)

            # and 1 different Tag to compare them
            meta_cucumber = sly.TagMeta('Cucumber', sly.TagValueType.ANY_STRING)
            tag_cucumber = sly.PointcloudTag(meta_cucumber, value="Fresh")

            # Compare identical Pointcloud Tags
            tag_lemon_1 == tag_lemon_2      # True

            # Compare unidentical Pointcloud Tags
            tag_lemon_1 == tag_cucumber     # False
        """
        return super(PointcloudTag, self).__eq__(other)

    def __ne__(self, other: PointcloudTag) -> bool:
        """
        Checks that 2 Pointcloud Tags are opposite.

        :param other: PointcloudTag object.
        :type other: :class:`PointcloudTag<PointcloudTag>`
        :return: True if comparable objects are not equal, otherwise False
        :rtype: :class:`bool`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Let's create 2 identical Tags
            meta_lemon_1 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            tag_lemon_1 = sly.PointcloudTag(meta_lemon_1)

            meta_lemon_2 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            tag_lemon_2 = sly.PointcloudTag(meta_lemon_2)

            # and 1 different Tag to compare them
            meta_cucumber = sly.TagMeta('Cucumber', sly.TagValueType.ANY_STRING)
            tag_cucumber = sly.PointcloudTag(meta_cucumber, value="Fresh")

            # Compare identical Pointcloud Tags
            tag_lemon_1 != tag_lemon_2      # False

            # Compare unidentical Pointcloud Tags
            tag_lemon_1 != tag_cucumber     # True
        """
        return super(PointcloudTag, self).__ne__(other)

    def key(self) -> uuid.UUID:
        return self._key

    def to_json(self, key_id_map: Optional[KeyIdMap] = None) -> Dict:
        """
        Convert the PointcloudTag to a json dict. 
        Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`KeyIdMap<supervisely.video_annotation.key_id_map.KeyIdMap>`, optional
        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
            tag_dog = sly.PointcloudTag(meta_dog)
            tag_dog_json = tag_dog.to_json()
            print(tag_dog_json)
            # Output: {
            #     "name": "dog",
            #     "key": "058ad7993a534082b4d94cc52542a97d"
            # }
        """
        data_json = super(PointcloudTag, self).to_json()
        if type(data_json) is str:
            # case when tag has no value, super.to_json() returns tag name
            data_json = {TagJsonFields.TAG_NAME: data_json}
        data_json[KEY] = self.key().hex

        if key_id_map is not None:
            item_id = key_id_map.get_tag_id(self.key())
            if item_id is not None:
                data_json[ID] = item_id

        return data_json

    @classmethod
    def from_json(cls, data: Dict, tag_meta_collection: TagMetaCollection, key_id_map: Optional[KeyIdMap] = None) -> PointcloudTag:
        """
        Convert a json dict to VideoTag. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: PointcloudTag in json format as a dict.
        :type data: :class:`dict`
        :param tag_meta_collection: TagMetaCollection object.
        :type tag_meta_collection: :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>`: TagMetaCollection
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`KeyIdMap<supervisely.video_annotation.key_id_map.KeyIdMap>`, optional
        :return: PointcloudTag object
        :rtype: :class:`PointcloudTag<PointcloudTag>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            tag_cat_json = {
                "name": "cat",
                "value": "Fluffy"
            }

            meta_cat = sly.TagMeta('cat', sly.TagValueType.ANY_STRING)
            meta_collection = sly.TagMetaCollection([meta_cat])
            tag_cat = sly.PointcloudTag.from_json(tag_cat_json, meta_collection)
        """
        temp = super(PointcloudTag, cls).from_json(data, tag_meta_collection)
        key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_tag(key, data.get(ID, None))

        return cls(meta=temp.meta, value=temp.value, key=key, sly_id=temp.sly_id, 
                   labeler_login=temp.labeler_login, updated_at=temp.updated_at, created_at=temp.created_at)

    def clone(
        self,
        meta: Optional[TagMeta] = None,
        value: Optional[Union[str, int, float]] = None,
        key: Optional[uuid.UUID] = None, 
        sly_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> PointcloudTag:
        """
        Clone makes a copy of Pointcloud Tag with new fields, if fields are given, otherwise it will use original Tag fields.

        :param meta: TagMeta object.
        :type meta: :class:`TagMeta<supervisely.annotation.tag_meta.TagMeta>`, optional
        :param value: Pointcloud Tag value. Depends on :class:`TagValueType<supervisely.annotation.tag_meta.TagValueType>` of :class:`TagMeta<supervisely.annotation.tag_meta.TagMeta>`.
        :type value: :class:`str` or :class:`int` or :class:`float` or :class:`NoneType`, optional
        :param key: uuid.UUID object.
        :type key: uuid.UUID, optional
        :param sly_id: Pointcloud Tag ID in Supervisely server.
        :type sly_id: :class:`int`, optional
        :param labeler_login: Login of user who created Pointcloud Tag.
        :type labeler_login: :class:`str`, optional
        :param updated_at: Date and Time when Pointcloud Tag was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: :class:`str`, optional
        :param created_at: Date and Time when Pointcloud Tag was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: :class:`str`, optional
        :return: New instance of Pointcloud Tag
        :rtype: :class:`PointcloudTag<PointcloudTag>`    
        :raises: :class:`ValueError`, If PointcloudTag value is incompatible to :class:`TagMeta<supervisely.annotation.tag_meta.TagMeta>` value type.
        
        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            # Original Pointcloud Tag
            weather_conditions = ["Sunny", "Cloudy", "Snowy", "Foggy", "Rainy"]
            meta_weather = sly.TagMeta("weather", sly.TagValueType.ONEOF_STRING, possible_values=weather_conditions)

            tag_weather = sly.PointcloudTag(meta_weather, value="Sunny")

            # Let's create some more tags by cloning our original Pointcloud Tag
            # Remember that PointcloudTag class object is immutable, and we need to assign new instance of PointcloudTag to a new variable
            clone_weather_1 = tag_weather.clone(value="Snowy")

            clone_weather_2 = tag_weather.clone(value="Cloudy")

            clone_weather_3 = tag_weather.clone(value="Rainy")
        """
        return PointcloudTag(
            meta=take_with_default(meta, self.meta),
            value=take_with_default(value, self.value),
            key=take_with_default(key, self.key()),
            sly_id=take_with_default(sly_id, self.sly_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
        )


# coding: utf-8

import math
import supervisely as sly


class ChunkSplitter:
    def __init__(self, tot_size, chunk_size):
        self.tot_size = tot_size
        self.chunk_size = chunk_size

    def __next__(self):
        for curr_pos in range(0, self.tot_size, self.chunk_size):
            curr_chunk_size = min(self.tot_size - curr_pos, self.chunk_size)
            yield (curr_pos, curr_chunk_size)

    def __iter__(self):
        return next(self)

    @property
    def chunk_cnt(self):
        return math.ceil(self.tot_size / self.chunk_size)


class ChunkedFileWriter:
    def __init__(self, file_path):
        self.handler = None
        self.path = file_path

        self.total_size = 0
        self.written_bytes = 0

    @property
    def file_path(self):
        return self.path

    # with proto chunks
    def write(self, chunk):
        if chunk.total_size != 0:
            self.total_size = chunk.total_size
        if self.written_bytes == 0:
            sly.fs.ensure_base_path(self.path)
            self.handler = open(self.path, 'wb')

        self.handler.write(chunk.buffer)
        self.written_bytes += len(chunk.buffer)

    # remove file on error
    def close_and_check(self):
        if (self.handler is not None) and (not self.handler.closed):
            self.handler.close()
        if (self.total_size != 0) and (self.written_bytes != self.total_size):
            sly.fs.silent_remove(self.path)
            return False
        return True


class ChunkedFileReader:
    def __init__(self, fpath, chunk_size):
        self.fpath = fpath
        self.file_size = sly.fs.get_file_size(fpath)  # bytes
        self.splitter = ChunkSplitter(self.file_size, chunk_size)

    def __next__(self):
        with open(self.fpath, 'rb') as file_:
            for chunk_start, chunk_size in self.splitter:
                chunk_bytes = file_.read(chunk_size)
                yield chunk_bytes

    def __iter__(self):
        return next(self)


def load_to_memory_chunked(iterable_resp):
    total_size = None
    b_data = bytearray()
    for chunk in iterable_resp:
        b_data.extend(chunk.buffer)
        if chunk.total_size != 0:
            total_size = chunk.total_size  # last non-zero value

    b_data_size = len(b_data)
    if (total_size is not None) and (total_size != b_data_size):
        raise RuntimeError('Incomplete input stream (by total_size). Expected: {}, Received: {}.'
                           .format(total_size, b_data_size))
    return b_data


def load_to_memory_chunked_image(iterable_resp):
    def iter_chunk_image(r):
        for chunk_image in r:
            yield chunk_image.chunk

    res = load_to_memory_chunked(iter_chunk_image(iterable_resp))
    return res



import copy
from typing import List
import cv2
import imutils
import numpy as np
import supervisely as sly
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from supervisely.app.widgets_context import JinjaWidgets


class ImageRegionSelector(Widget):
    class Routes:
        BBOX_CHANGED = "bbox_changed"
        POSITIVE_CHANGED = "positive-updated"
        NEGATIVE_CHANGED = "negative-updated"

    def __init__(
        self,
        image_info: sly.ImageInfo = None,
        mask: sly.Bitmap = None,
        mask_opacity: int = 50,
        bbox: List[int] = None,
        widget_id: str = None,
        disabled: bool = False,
        widget_width: str = "100%",
        widget_height: str = "100%",
    ):
        self._image_info = None
        self._image_link = None
        self._image_name = None
        self._image_url = None
        self._image_hash = None
        self._image_size = None
        self._image_id = None
        self._image_width = None
        self._image_height = None
        self._dataset_id = None
        self._mask = None
        self._mask_opacity = mask_opacity
        self._original_bbox = None
        self._scaled_bbox = None
        self._disabled = disabled
        self._widget_width = widget_width
        self._widget_height = widget_height
        self._changes_handled = False
        self._pos_points_changes_handled = False
        self._neg_points_changes_handled = False

        super().__init__(widget_id=widget_id, file_path=__file__)
        if image_info is not None:
            self.image_update(image_info)
        
        if mask is not None:
            self.set_mask(mask)
        
        if bbox is not None:
            self.set_bbox(bbox)

        script_path = "./sly/css/app/widgets/image_region_selector/script.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__ + "1"] = "https://cdn.jsdelivr.net/npm/svg.js@2.7.1/dist/svg.min.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__ + "2"] = "https://cdn.jsdelivr.net/npm/svg.select.js@3.0.1/dist/svg.select.min.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__ + "3"] = "https://cdn.jsdelivr.net/npm/svg.resize.js@1.4.3/dist/svg.resize.min.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__ + "4"] = "https://cdn.jsdelivr.net/npm/svg.draggable.js@2.2.2/dist/svg.draggable.min.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__ + "5"] = "https://cdn.jsdelivr.net/npm/svg.panzoom.js@1.2.3/dist/svg.panzoom.min.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__ + "6"] = "https://rawgit.com/nodeca/pako/1.0.11/dist/pako.min.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__ + "7"] = "https://cdnjs.cloudflare.com/ajax/libs/uuid/8.3.2/uuidv4.min.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__] = script_path

    def image_update(self, image_info: sly.ImageInfo):
        self._image_info = image_info
        self._image_link = image_info.preview_url
        self._image_name = image_info.name
        self._image_url = image_info.preview_url
        self._image_hash = image_info.hash
        self._image_size = image_info.size
        self._image_id = image_info.id
        self._image_width = image_info.width
        self._image_height = image_info.height
        self._dataset_id = image_info.dataset_id

        padding_pix = int(max([image_info.width, image_info.height]) * 0.1)
        self._original_bbox = [
            [padding_pix, padding_pix],
            [image_info.width - padding_pix, image_info.height - padding_pix],
        ]
        self._scaled_bbox = self._original_bbox

        StateJson()[self.widget_id].update(self.get_json_state())
        StateJson().send_changes()
    
    def set_image(self, image_info: sly.ImageInfo):
        self._image_info = image_info
        self._image_link = image_info.preview_url
        self._image_name = image_info.name
        self._image_url = image_info.preview_url
        self._image_hash = image_info.hash
        self._image_size = image_info.size
        self._image_id = image_info.id
        self._image_width = image_info.width
        self._image_height = image_info.height
        self._dataset_id = image_info.dataset_id
        
        padding_pix = int(max([image_info.width, image_info.height]) * 0.1)
        self._original_bbox = [
            [padding_pix, padding_pix],
            [image_info.width - padding_pix, image_info.height - padding_pix],
        ]
        self._scaled_bbox = self._original_bbox
        
        StateJson()[self.widget_id].update(self.get_json_state())
        StateJson().send_changes()

    def get_image_info(self):
        return self._image_info

    def set_bbox(self, bbox):
        if bbox is None:
            padding_pix = int(max([self._image_width, self._image_height]) * 0.1)
            bbox = [
                [padding_pix, padding_pix],
                [self._image_width - padding_pix, self._image_height - padding_pix],
            ]

        self._original_bbox = bbox
        self._scaled_bbox = self._original_bbox

        StateJson()[self.widget_id].update(self.get_json_state())
        StateJson().send_changes()

    def get_bbox(self):
        return self._original_bbox

    def set_mask(self, bitmap: sly.Bitmap, color="#77e377"):
        mask = None
        if bitmap is not None:
            mask_base64 = bitmap.data_2_base64(bitmap.data)
            origin = [bitmap.origin.col, bitmap.origin.row]
            contours = copy.deepcopy(self._get_contours(mask_base64, origin))
            mask = {
                'data': mask_base64,
                'origin': origin,
                'color': color,
                'contour': contours
            }
        self._mask = mask
        self.update_state()
        StateJson().send_changes()

    def get_mask(self):
        return self._mask

    def bbox_changed(self, func, page_path=""):
        route_path = page_path + self.get_route_path(ImageRegionSelector.Routes.BBOX_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            self.bbox_update()
            res = self._scaled_bbox
            func(res)

        return _click
    
    def positive_points_changed(self, func, page_path=""):
        route_path = page_path + self.get_route_path(ImageRegionSelector.Routes.POSITIVE_CHANGED)
        server = self._sly_app.get_server()
        self._pos_points_changes_handled = True

        @server.post(route_path)
        def _click():
            points = self.get_positive_points()
            func(points)

        return _click
    
    def negative_points_changed(self, func, page_path=""):
        route_path = page_path + self.get_route_path(ImageRegionSelector.Routes.NEGATIVE_CHANGED)
        server = self._sly_app.get_server()
        self._neg_points_changes_handled = True

        @server.post(route_path)
        def _click():
            points = self.get_negative_points()
            func(points)

        return _click

    def bbox_update(self):
        self._scaled_bbox = StateJson()[self.widget_id]["scaledBbox"]
        bboxes_padding = 0

        scaled_width, scaled_height = self.get_bbox_size(self._scaled_bbox)
        original_width, original_height = int(scaled_width / (1 + bboxes_padding)), int(
            scaled_height / (1 + bboxes_padding)
        )

        div_width, div_height = (scaled_width - original_width) // 2, (
            scaled_height - original_height
        ) // 2

        self._original_bbox[0][0] = self._scaled_bbox[0][0] + div_width
        self._original_bbox[0][1] = self._scaled_bbox[0][1] + div_height
        self._original_bbox[1][0] = self._scaled_bbox[1][0] - div_width
        self._original_bbox[1][1] = self._scaled_bbox[1][1] - div_height
        StateJson().send_changes()

    def get_json_data(self):
        return {
            "disabled": self._disabled,
        }

    def get_json_state(self):
        return {
            "imageLink": self._image_link,
            "imageName": self._image_name,
            "imageUrl": self._image_url,
            "imageHash": self._image_hash,
            "imageSize": self._image_size,
            "imageId": self._image_id,
            "imageWidth": self._image_width,
            "imageHeight": self._image_height,
            "mask": self._mask,
            "mask_opacity": self._mask_opacity,
            "datasetId": self._dataset_id,
            "originalBbox": self._original_bbox,
            "scaledBbox": self._scaled_bbox,
            "disabled": self._disabled,
            "widget_width": self._widget_width,
            "widget_height": self._widget_height,
            "widget_id": self.widget_id,
        }

    @property
    def is_empty(self):
        return not self._image_info is None

    def get_bbox_size(self, current_bbox):
        box_width = current_bbox[1][0] - current_bbox[0][0]
        box_height = current_bbox[1][1] - current_bbox[0][1]
        return box_width, box_height

    def add_bbox_padding(self, padding_coefficient=0):
        padding_coefficient /= 100

        original_w, original_h = self.get_bbox_size(current_bbox=self._original_bbox)
        additional_w, additional_h = (
            int(original_w * padding_coefficient // 2),
            int(original_h * padding_coefficient // 2),
        )

        self._scaled_bbox[0][0] = (
            self._original_bbox[0][0] - additional_w
            if self._original_bbox[0][0] - additional_w > 0
            else 0
        )
        self._scaled_bbox[0][1] = (
            self._original_bbox[0][1] - additional_h
            if self._original_bbox[0][1] - additional_h > 0
            else 0
        )
        self._scaled_bbox[1][0] = (
            self._original_bbox[1][0] + additional_w
            if self._original_bbox[1][0] + additional_w < self._image_size[0]
            else self._image_size[0] - 1
        )
        self._scaled_bbox[1][1] = (
            self._original_bbox[1][1] + additional_h
            if self._original_bbox[1][1] + additional_h < self._image_size[1]
            else self._image_size[1] - 1
        )
        StateJson().send_changes()

    def get_relative_coordinates(self, abs_coordinates):
        box_width, box_height = self.get_bbox_size(current_bbox=self._scaled_bbox)
        return {
            "x": (abs_coordinates["position"][0][0] - self._scaled_bbox[0][0]) / box_width,
            "y": (abs_coordinates["position"][0][1] - self._scaled_bbox[0][1]) / box_height,
        }

    def _get_contours(self, base64mask, origin_shift):
        test_mask = np.asarray(sly.Bitmap.base64_2_data(base64mask)).astype(np.uint8) * 255
        thresholded_mask = cv2.threshold(test_mask, 100, 255, cv2.THRESH_BINARY)[1]

        contours = cv2.findContours(thresholded_mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        contours = imutils.grab_contours(contours)

        reshaped_contours = []
        for contour in contours:
            reshaped_contour = contour.reshape(contour.shape[0], 2)

            reshaped_contour[:, 0] += origin_shift[0]
            reshaped_contour[:, 1] += origin_shift[1]

            reshaped_contours.append(reshaped_contour.tolist())

        return reshaped_contours
    
    def get_positive_points(self):
        return StateJson()[self.widget_id]["positivePoints"]

    def get_negative_points(self):
        return StateJson()[self.widget_id]["negativePoints"]



from supervisely.app.widgets import Text, Select, Menu, Field, Container, Button
from supervisely.imaging.color import hex2rgb
from supervisely.nn.inference.instance_segmentation.dashboard.deploy_ui import (
    deploy_layout,
)
from supervisely.nn.inference.instance_segmentation.dashboard.classes_ui import (
    classes_layout,
)
from supervisely.nn.inference.instance_segmentation.dashboard.preview_image import (
    preview_image_layout,
)

l = Text(text="left part", status="success")


ttt = Text(text="some text", status="warning")
# sidebar = sly.app.widgets.Sidebar(left_pane=l, right_pane=item)

g1 = Menu.Group(
    "Model",
    [
        Menu.Item(
            title="Deployment / Run", content=deploy_layout, icon="zmdi zmdi-dns"
        ),
        Menu.Item(title="Classes", content=classes_layout, icon="zmdi zmdi-shape"),
        Menu.Item(title="Monitoring", content=l, icon="zmdi zmdi-chart"),
    ],
)
g2 = Menu.Group(
    "Preview predictions",
    [
        Menu.Item(title="Image", content=preview_image_layout, icon="zmdi zmdi-image"),
        Menu.Item(title="Video", content=ttt, icon="zmdi zmdi-youtube-play"),
    ],
)
g3 = Menu.Group(
    "Inference",
    [
        Menu.Item(
            title="Apply to images project",
            content=ttt,
            icon="zmdi zmdi-collection-folder-image",
        ),
        Menu.Item(
            title="Apply to videos project",
            content=ttt,
            icon="zmdi zmdi-collection-video",
        ),
    ],
)
menu = Menu(groups=[g1, g2, g3], index="Image")
# menu = sly.app.widgets.Menu(items=g1_items)



# coding: utf-8

# docs
from typing import List, Optional

from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.api.entity_annotation.object_api import ObjectApi


class VideoObjectApi(ObjectApi):
    """
    :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` for :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>`.
    """

    def append_bulk(
        self, video_id: int, objects: VideoObjectCollection, key_id_map: Optional[KeyIdMap] = None
    ) -> List[int]:
        """
        Add Tags to Annotation Objects

        :param video_id: Video ID in Supervidely.
        :type video_id: int
        :param objects: VideoAnnotation objects.
        :type objects: VideoObjectCollection
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: List of objects IDs
        :rtype: :class:`List[int]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            from supervisely.video_annotation.video_tag import VideoTag
            from supervisely.video_annotation.key_id_map import KeyIdMap

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_id = 17209
            video_id = 19402023

            meta_json = api.project.get_meta(project_id)
            project_meta = sly.ProjectMeta.from_json(meta_json)

            key_id_map = KeyIdMap()
            ann_info = api.video.annotation.download(video_id)
            ann = sly.VideoAnnotation.from_json(ann_info, project_meta, key_id_map)

            api.video.object.append_bulk(video_id, ann.objects, key_id_map)
        """

        info = self._api.video.get_info_by_id(video_id)
        return self._append_bulk(
            self._api.video.tag, video_id, info.project_id, info.dataset_id, objects, key_id_map
        )



from __future__ import annotations
from typing import Optional, List, Dict

from supervisely._utils import take_with_default
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.pointcloud_annotation.pointcloud_object_collection import PointcloudObjectCollection

class PointcloudEpisodeFrame(Frame):
    figure_type = PointcloudFigure

    def __init__(self, index: int, figures: Optional[List[PointcloudFigure]]=None):
        super().__init__(index, figures)

    @classmethod
    def from_json(
        cls, 
        data: Dict, 
        objects: PointcloudObjectCollection, 
        frames_count: Optional[int]=None, 
        key_id_map: Optional[KeyIdMap]=None
    ) -> PointcloudEpisodeFrame:
        return super().from_json(data, objects, frames_count, key_id_map)

    def clone(self, index: Optional[int] = None, figures: Optional[List[PointcloudFigure]] = None) -> PointcloudEpisodeFrame:
        return super().clone(index, figures)